from django.urls import reverse


def emailSubjectPrefix():
    return "Code for Life"


def emailBodySignOff(request):
    return (
        f"\n\nThanks,\n\nThe Code for Life team.\n"
        f"{request.build_absolute_uri(reverse('home'))}"
    )


def emailVerificationNeededEmail(request, token):
    return {
        "subject": f"{emailSubjectPrefix()}: Email address verification needed",
        "message": (
            f"Please go to "
            f"{request.build_absolute_uri(reverse('verify_email', kwargs={'token': token}))} "
            f"to verify your email address."
            f"{emailBodySignOff(request)}"
        ),
    }


def emailChangeVerificationEmail(request, token):
    return {
        "subject": f"{emailSubjectPrefix()}: Email address verification needed",
        "message": (
            f"You are changing your email, please go to "
            f"{request.build_absolute_uri(reverse('verify_email', kwargs={'token': token}))} "
            f"to verify your new email address. If you are not part of Code for Life "
            f"then please ignore this email."
            f"{emailBodySignOff(request)}"
        ),
    }


def emailChangeNotificationEmail(request):
    return {
        "subject": f"{emailSubjectPrefix()}: Email address changed",
        "message": (
            f"Someone has tried to change the email address of your account. If this "
            f"was not you, please get in contact with us."
            f"{emailBodySignOff(request)}"
        ),
    }


def emailChangeDuplicateNotificationEmail(request, email):
    return {
        "subject": f"{emailSubjectPrefix()}: Duplicate account error",
        "message": (
            f"A user is already registered with this email address: {email}.\n"
            f"Please change your email address to something else."
            f"{emailBodySignOff(request)}"
        ),
    }


def userAlreadyRegisteredEmail(request, email, is_independent_student=False):
    if is_independent_student:
        login_url = reverse("independent_student_login")
    else:
        login_url = reverse("teacher_login")

    return {
        "subject": f"{emailSubjectPrefix()}: Duplicate account error",
        "message": (
            f"A user is already registered with this email address: {email}.\n"
            f"If you've already registered, please login: "
            f"{request.build_absolute_uri(login_url)}.\n"
            f"Otherwise please register with a different email address."
            f"{emailBodySignOff(request)}"
        ),
    }


def indepStudentUsernameAlreadyExistsEmail(request, username):
    return {
        "subject": f"{emailSubjectPrefix()}: Username already taken",
        "message": f"A user is already registered with this username: {username}.\n"
        f"If you've already registered, please login: "
        f"{request.build_absolute_uri(reverse('independent_student_login'))}.\n"
        f"Otherwise please register with a different username."
        f"{emailBodySignOff(request)}",
    }


def joinRequestPendingEmail(request, pendingAddress):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request pending",
        "message": (
            f"Someone with the email address '{pendingAddress}' has asked to join your "
            f"school or club, please go to "
            f"{request.build_absolute_uri(reverse('dashboard'))} to view the pending "
            f"join request."
            f"{emailBodySignOff(request)}"
        ),
    }


def joinRequestSentEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request sent",
        "message": (
            f"Your request to join the school or club '{schoolName}' has been sent. "
            f"Someone will either accept or deny your request soon."
            f"{emailBodySignOff(request)}"
        ),
    }


def joinRequestAcceptedEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request accepted",
        "message": (
            f"Your request to join the school or club '{schoolName}' has been accepted."
            f"{emailBodySignOff(request)}"
        ),
    }


def joinRequestDeniedEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request denied",
        "message": (
            f"Your request to join the school or club '{schoolName}' has been denied. "
            f"If you think this was in error you should speak to the administrator of "
            f"that school or club."
            f"{emailBodySignOff(request)}"
        ),
    }


def kickedEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: You were removed from your school or club",
        "message": (
            f"You have been removed from the school or club '{schoolName}'. "
            f"If you think this was an error, please contact the administrator of that "
            f"school or club."
            f"{emailBodySignOff(request)}"
        ),
    }


def adminGivenEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: You have been made a school or club administrator",
        "message": (
            f"Administrator control of the school or club '{schoolName}' has been "
            f"given to you. Go to {request.build_absolute_uri(reverse('dashboard'))} "
            f"to start managing your school or club."
            f"{emailBodySignOff(request)}"
        ),
    }


def adminRevokedEmail(request, schoolName):
    return {
        "subject": f"{emailSubjectPrefix()}: You are no longer a school or club administrator",
        "message": (
            f"Your administrator control of the school or club '{schoolName}' has been "
            f"revoked. If you think this is an error, please contact one of the other "
            f"administrators in your school or club."
            f"{emailBodySignOff(request)}"
        ),
    }


def studentJoinRequestSentEmail(request, schoolName, accessCode):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request sent",
        "message": (
            f"Your request to join the school or club '{schoolName}' in class "
            f"{accessCode} has been sent to that class's teacher, who will either "
            f"accept or deny your request."
            f"{emailBodySignOff(request)}"
        ),
    }


def studentJoinRequestNotifyEmail(request, username, email, accessCode):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request by student {username}",
        "message": (
            f"There is a request waiting from student with username '{username}' and "
            f"email {email} to join your class {accessCode}. Go to "
            f"{request.build_absolute_uri(reverse('dashboard'))} to review the request."
            f"{emailBodySignOff(request)}"
        ),
    }


def studentJoinRequestRejectedEmail(request, schoolName, accessCode):
    return {
        "subject": f"{emailSubjectPrefix()}: School or club join request rejected",
        "message": (
            f"Your request to join the school or club '{schoolName}' in class "
            f"{accessCode} has been rejected. Speak to your teacher if you think this "
            f"is an error."
            f"{emailBodySignOff(request)}"
        ),
    }


def inviteTeacherEmail(request):
    return {
        "subject": f"{emailSubjectPrefix()}: You've been invited to join Code for Life",
        "message": (
            f"A colleague at your school or code club has invited you to become part of "
            f"Code for Life.\n\nPlease register your details to get started.\n\n"
            f"{request.build_absolute_uri(reverse('register'))}"
            f"{emailBodySignOff(request)}"
        ),
    }



from common.models import Student, Teacher, School
from django.test import TestCase
from django.utils import timezone

from .utils.classes import create_class_directly
from .utils.organisation import create_organisation_directly
from .utils.student import create_independent_student_directly
from .utils.teacher import signup_teacher_directly


class TestModels(TestCase):
    def test_indep_student_pending_class_request_on_delete(self):
        """
        Given a class and an independent student,
        When the student makes a request to join the class and the class is deleted,
        Then the student's pending class request field is set to null.
        """
        teacher_email, _ = signup_teacher_directly()
        school_name, _ = create_organisation_directly(teacher_email)
        class_name = "Test Class"
        klass, _, _ = create_class_directly(teacher_email, class_name)

        username, _, indep_student = create_independent_student_directly()

        assert indep_student.is_independent()
        assert indep_student.pending_class_request is None

        indep_student.pending_class_request = klass
        indep_student.save()

        klass.delete()

        indep_student = Student.objects.get(new_user__username=username)

        assert indep_student.pending_class_request is None

    def test_teacher_school_on_delete(self):
        """
        Given a school and a teacher in that school,
        When the school is deleted,
        Then the teacher's school field is set to null.
        """
        teacher_email, _ = signup_teacher_directly()
        school_name, _ = create_organisation_directly(teacher_email)

        teacher = Teacher.objects.get(new_user__email=teacher_email)
        school = School.objects.get(name=school_name)

        assert teacher.school == school

        school.delete()
        teacher = Teacher.objects.get(new_user__email=teacher_email)

        assert teacher.school is None

    def test_teacher_pending_join_request_on_delete(self):
        """
        Given a school and a teacher without a school,
        When the teacher requests to join the school, and that school is deleted,
        Then the teacher's pending join request field is set to null.
        """
        teacher1_email, _ = signup_teacher_directly()
        teacher2_email, _ = signup_teacher_directly()
        school_name, _ = create_organisation_directly(teacher1_email)

        teacher1 = Teacher.objects.get(new_user__email=teacher1_email)
        teacher2 = Teacher.objects.get(new_user__email=teacher2_email)
        school = School.objects.get(name=school_name)

        assert teacher1.school == school
        assert teacher2.school is None
        assert teacher2.pending_join_request is None

        teacher2.pending_join_request = school
        teacher2.save()

        school.delete()

        teacher2 = Teacher.objects.get(new_user__email=teacher2_email)

        assert teacher2.school is None
        assert teacher2.pending_join_request is None

    def test_creation_time(self):
        teacher_email, _ = signup_teacher_directly()

        sometime = timezone.now()  # mark time before the school creation
        school_name, _ = create_organisation_directly(teacher_email)

        school = School.objects.get(name=school_name)
        # check the creation time
        assert school.creation_time > sometime

        sometime = timezone.now()  # mark time before the class creation
        klass, name, access_code = create_class_directly(teacher_email)
        # check the creation time
        assert klass.creation_time > sometime



# Generated by Django 2.2.24 on 2021-10-15 15:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0012_usersession'),
    ]

    operations = [
        migrations.AddField(
            model_name='class',
            name='creation_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='school',
            name='creation_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='usersession',
            name='class_field',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='common.Class'),
        ),
        migrations.AddField(
            model_name='usersession',
            name='school',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='common.School'),
        ),
    ]



import datetime
import json
from uuid import uuid4

from common import app_settings
from common.email_messages import (
    emailChangeNotificationEmail,
    emailChangeVerificationEmail,
    emailChangeDuplicateNotificationEmail,
    emailVerificationNeededEmail,
)
from common.models import EmailVerification, Teacher, Student
from django.core.mail import EmailMultiAlternatives
from django.http import HttpResponse
from django.template import loader
from django.utils import timezone
from requests import post, get, put
from requests.exceptions import RequestException

NOTIFICATION_EMAIL = "Code For Life Notification <" + app_settings.EMAIL_ADDRESS + ">"
VERIFICATION_EMAIL = "Code For Life Verification <" + app_settings.EMAIL_ADDRESS + ">"
PASSWORD_RESET_EMAIL = (
    "Code For Life Password Reset <" + app_settings.EMAIL_ADDRESS + ">"
)
INVITE_FROM = "Code For Life Invitation <" + app_settings.EMAIL_ADDRESS + ">"


def send_email(
    sender,
    recipients,
    subject,
    text_content,
    html_content=None,
    plaintext_template="email.txt",
    html_template="email.html",
):
    # add in template for templates to message

    # setup templates
    plaintext = loader.get_template(plaintext_template)
    html = loader.get_template(html_template)
    plaintext_email_context = {"content": text_content}
    html_email_context = {"content": text_content}
    if html_content:
        html_email_context = {"content": html_content}

    # render templates
    plaintext_body = plaintext.render(plaintext_email_context)
    html_body = html.render(html_email_context)

    # make message using templates
    message = EmailMultiAlternatives(subject, plaintext_body, sender, recipients)
    message.attach_alternative(html_body, "text/html")

    message.send()


def generate_token(user, email="", preverified=False):
    return EmailVerification.objects.create(
        user=user,
        email=email,
        token=uuid4().hex[:30],
        expiry=timezone.now() + datetime.timedelta(hours=1),
        verified=preverified,
    )


def send_verification_email(request, user, new_email=None):
    """Send an email prompting the user to verify their email address."""

    if not new_email:  # verifying first email address
        user.email_verifications.all().delete()

        verification = generate_token(user)

        message = emailVerificationNeededEmail(request, verification.token)
        send_email(
            VERIFICATION_EMAIL, [user.email], message["subject"], message["message"]
        )

    else:  # verifying change of email address.
        verification = generate_token(user, new_email)

        message = emailChangeVerificationEmail(request, verification.token)
        send_email(
            VERIFICATION_EMAIL, [new_email], message["subject"], message["message"]
        )

        message = emailChangeNotificationEmail(request)
        send_email(
            VERIFICATION_EMAIL, [user.email], message["subject"], message["message"]
        )


def is_verified(user):
    """Check that a user has verified their email address."""
    verifications = user.email_verifications.filter(verified=True)
    return len(verifications) != 0


def add_to_dotmailer(first_name: str, last_name: str, email: str):
    try:
        create_contact(first_name, last_name, email)
        add_contact_to_address_book(first_name, last_name, email)
    except RequestException:
        return HttpResponse(status=404)


def create_contact(first_name, last_name, email):
    url = app_settings.DOTMAILER_CREATE_CONTACT_URL
    body = {
        "contact": {
            "email": email,
            "optInType": "VerifiedDouble",
            "emailType": "Html",
            "dataFields": [
                {"key": "FIRSTNAME", "value": first_name},
                {"key": "LASTNAME", "value": last_name},
                {"key": "FULLNAME", "value": f"{first_name} {last_name}"},
            ],
        },
        "consentFields": [
            {
                "fields": [
                    {
                        "key": "DATETIMECONSENTED",
                        "value": datetime.datetime.now().__str__(),
                    },
                ]
            }
        ],
        "preferences": app_settings.DOTMAILER_DEFAULT_PREFERENCES,
    }

    post(
        url,
        json=body,
        auth=(app_settings.DOTMAILER_USER, app_settings.DOTMAILER_PASSWORD),
    )


def add_contact_to_address_book(first_name, last_name, email):
    url = app_settings.DOTMAILER_ADDRESS_BOOK_URL
    body = {
        "email": email,
        "optInType": "VerifiedDouble",
        "emailType": "Html",
        "dataFields": [
            {"key": "FIRSTNAME", "value": first_name},
            {"key": "LASTNAME", "value": last_name},
            {"key": "FULLNAME", "value": f"{first_name} {last_name}"},
        ],
    }

    post(
        url,
        json=body,
        auth=(app_settings.DOTMAILER_USER, app_settings.DOTMAILER_PASSWORD),
    )


def get_dotmailer_user_by_email(email):
    url = app_settings.DOTMAILER_GET_USER_BY_EMAIL_URL.replace("EMAIL", email)

    response = get(
        url,
        auth=(app_settings.DOTMAILER_USER, app_settings.DOTMAILER_PASSWORD),
    )

    return json.loads(response.content)


def add_consent_record_to_dotmailer_user(user):
    consent_date_time = datetime.datetime.now().__str__()

    url = app_settings.DOTMAILER_PUT_CONSENT_DATA_URL.replace(
        "USER_ID", str(user["id"])
    )
    body = {
        "contact": {
            "email": user["email"],
            "optInType": user["optInType"],
            "emailType": user["emailType"],
            "dataFields": user["dataFields"],
        },
        "consentFields": [
            {
                "fields": [
                    {"key": "DATETIMECONSENTED", "value": consent_date_time},
                ]
            }
        ],
    }

    put(
        url,
        json=body,
        auth=(app_settings.DOTMAILER_USER, app_settings.DOTMAILER_PASSWORD),
    )


def send_dotmailer_consent_confirmation_email_to_user(user):
    url = app_settings.DOTMAILER_SEND_CAMPAIGN_URL
    campaign_id = app_settings.DOTMAILER_THANKS_FOR_STAYING_CAMPAIGN_ID
    body = {
        "campaignID": campaign_id,
        "contactIds": [str(user["id"])],
    }

    post(
        url,
        json=body,
        auth=(app_settings.DOTMAILER_USER, app_settings.DOTMAILER_PASSWORD),
    )


def update_email(user: Teacher or Student, request, data):
    changing_email = False
    new_email = data["email"]

    if new_email != "" and new_email != user.new_user.email:
        changing_email = True
        if _is_email_already_taken(new_email, user):
            email_message = emailChangeDuplicateNotificationEmail(request, new_email)
            send_email(
                NOTIFICATION_EMAIL,
                [user.new_user.email],
                email_message["subject"],
                email_message["message"],
            )
        else:
            # new email to set and verify
            send_verification_email(request, user.new_user, new_email)
    return changing_email, new_email


def _is_email_already_taken(new_email, user):
    teachers_with_email = Teacher.objects.filter(new_user__email=new_email)
    students_with_email = Student.objects.filter(new_user__email=new_email)

    return (teachers_with_email.exists() and teachers_with_email[0] != user) or (
        students_with_email.exists() and students_with_email[0] != user
    )
